Module: Opal::UseGem

Included in:
Opal, Builder
Defined in:
lib/opal/paths.rb

Instance Method Summary collapse

Instance Method Details

#use_gem(gem_name, include_dependencies = true) ⇒ Object

Adds the "require_paths" (usually lib/) of gem with the given name to Opal paths. By default will include the "require_paths" from all the dependent gems.

Parameters:

  • the name of the gem

  • (defaults to: true)

    whether or not to add recursively the gem's dependencies

Raises:

  • if gem or any of its runtime dependencies not found



57
58
59
# File 'lib/opal/paths.rb', line 57

def use_gem(gem_name, include_dependencies = true)
  append_paths(*require_paths_for_gem(gem_name, include_dependencies))
end