Method: Opal::Compiler#parser_indent
- Defined in:
- lib/opal/compiler.rb
#parser_indent ⇒ Object
Instances of Scope can use this to determine the current
scope indent. The indent is used to keep generated code easily
readable.
386 387 388 |
# File 'lib/opal/compiler.rb', line 386 def parser_indent @indent end |